from unittest import TestCase, mock

from pcs.common import const, reports
from pcs.lib.commands.constraint import ticket as ticket_command

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import create_patcher

patch_commands = create_patcher("pcs.lib.commands.constraint.ticket")


class CreateTest(TestCase):
    def test_success_create_minimal(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <primitive id="resourceA" class="service" type="exim"/>
                </resources>
            """,
        )
        config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket-ticketA-resourceA"
                        rsc="resourceA" ticket="ticketA"
                    />
                </constraints>
            """
        )

        ticket_command.create(env_assist.get_env(), "ticketA", "resourceA", {})

    def test_success_create_full(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <primitive id="resourceA" class="service" type="exim"/>
                </resources>
            """,
        )
        config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="my-id" rsc="resourceA" rsc-role="{role}"
                        ticket="ticketA" loss-policy="fence"
                    />
                </constraints>
            """.format(role=const.PCMK_ROLE_PROMOTED)
        )

        ticket_command.create(
            env_assist.get_env(),
            "ticketA",
            "resourceA",
            {
                "loss-policy": "Fence",
                "rsc-role": str(const.PCMK_ROLE_PROMOTED).lower(),
                "id": "my-id",
            },
        )

    def test_refuse_legacy_role(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                    <resources>
                        <primitive id="resourceA" class="service" type="exim"/>
                    </resources>
                """,
        )

        role = str(const.PCMK_ROLE_UNPROMOTED_LEGACY).lower()
        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(
                env_assist.get_env(),
                "ticketA",
                "resourceA",
                {"rsc-role": role},
            ),
        )
        env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=role,
                    allowed_values=const.PCMK_ROLES,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_refuse_for_nonexisting_resource(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load()
        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(
                env_assist.get_env(),
                "ticketA",
                "resourceA",
                {},
            ),
        )
        env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="resourceA",
                    expected_types=[],
                    context_type="",
                    context_id="",
                )
            ]
        )

    def test_validation(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <primitive id="R" class="service" type="exim">
                        <meta_attributes id="R-meta"/>
                    </primitive>
                </resources>
            """,
        )
        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(
                env_assist.get_env(),
                "T_1",
                "R-meta",
                {
                    "loss-policy": "lossX",
                    "rsc-role": "roleX",
                    "id": "R",
                    "bad": "option",
                },
            )
        )
        env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    reports.codes.BOOTH_TICKET_NAME_INVALID,
                    ticket_name="T_1",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["bad"],
                    allowed=["id", "loss-policy", "rsc-role"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="R",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value="roleX",
                    allowed_values=(
                        const.PCMK_ROLE_STOPPED,
                        const.PCMK_ROLE_STARTED,
                        const.PCMK_ROLE_PROMOTED,
                        const.PCMK_ROLE_UNPROMOTED,
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="loss-policy",
                    option_value="lossX",
                    allowed_values=("fence", "stop", "freeze", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_duplicate_constraint(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <primitive id="R" class="service" type="exim"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_ticket id="ticket-T-R" rsc="R" ticket="T" />
                </constraints>
            """,
        )

        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(env_assist.get_env(), "T", "R", {})
        )
        env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    force_code=reports.codes.FORCE,
                    constraint_ids=["ticket-T-R"],
                )
            ]
        )

    def test_duplicate_constraint_forced(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <primitive id="R" class="service" type="exim" />
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_ticket id="ticket-T-R" rsc="R" ticket="T" />
                </constraints>
            """,
        )
        config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket-T-R" rsc="R" ticket="T" />
                    <rsc_ticket id="ticket-T-R-1" rsc="R" ticket="T" />
                </constraints>
            """
        )

        ticket_command.create(
            env_assist.get_env(), "T", "R", {}, duplication_alowed=True
        )
        env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    constraint_ids=["ticket-T-R"],
                )
            ]
        )

    def test_resource_in_clone(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <clone id="C">
                        <primitive id="R" class="service" type="exim" />
                    </clone>
                </resources>
            """,
        )

        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(env_assist.get_env(), "T", "R", {})
        )

        env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="R",
                    parent_type="clone",
                    parent_id="C",
                )
            ]
        )

    def test_resource_in_clone_forced(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources="""
                <resources>
                    <clone id="C">
                        <primitive id="R" class="service" type="exim" />
                    </clone>
                </resources>
            """,
        )
        config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket-T-R" rsc="R" ticket="T" />
                </constraints>
            """
        )

        ticket_command.create(
            env_assist.get_env(), "T", "R", {}, resource_in_clone_alowed=True
        )

        env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="R",
                    parent_type="clone",
                    parent_id="C",
                )
            ]
        )


class CreateWithSet(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.cib_resources = """
            <resources>
                <primitive id="resource1">
                    <meta_attributes id="resource1-meta"/>
                </primitive>
                <primitive id="resource2">
                    <meta_attributes id="resource2-meta"/>
                </primitive>
                <primitive id="resource3" />
                <primitive id="resource4" />
                <clone id="C">
                    <primitive id="resource-in-clone" />
                </clone>
                <bundle id="B">
                    <primitive id="resource-in-bundle" />
                </bundle>
            </resources>
        """

    @staticmethod
    def set(*args, **kwargs):
        if "require_all" in kwargs:
            kwargs["require-all"] = kwargs["require_all"]
            del kwargs["require_all"]
        return dict(ids=args, options=kwargs)

    def test_success_create_minimal(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket_set_r1" ticket="ticket1">
                        <resource_set id="ticket_set_r1_set">
                            <resource_ref id="resource1" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """
        )
        ticket_command.create_with_set(
            self.env_assist.get_env(),
            [self.set("resource1")],
            {"ticket": "ticket1"},
        )

    def test_success_create_full(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="constr1" loss-policy="freeze"
                        ticket="ticket1"
                    >
                        <resource_set id="constr1_set" action="promote"
                            require-all="true"
                        >
                            <resource_ref id="resource1" />
                            <resource_ref id="resource2" />
                        </resource_set>
                        <resource_set id="constr1_set-1" role="{role}"
                            sequential="true"
                        >
                            <resource_ref id="resource3" />
                            <resource_ref id="resource4" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """.format(role=const.PCMK_ROLE_PROMOTED)
        )
        ticket_command.create_with_set(
            self.env_assist.get_env(),
            [
                self.set(
                    "resource1",
                    "resource2",
                    action="promote",
                    require_all="true",
                ),
                self.set(
                    "resource3",
                    "resource4",
                    role=const.PCMK_ROLE_PROMOTED,
                    sequential="true",
                ),
            ],
            {"ticket": "ticket1", "loss-policy": "FReeZE", "id": "constr1"},
        )

    def test_refuse_legacy_role(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        role = str(const.PCMK_ROLE_UNPROMOTED_LEGACY).lower()
        self.env_assist.assert_raise_library_error(
            lambda: ticket_command.create_with_set(
                self.env_assist.get_env(),
                [self.set("resource1", role=role)],
                {"ticket": "ticket1"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value=role,
                    allowed_values=const.PCMK_ROLES,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_refuse_for_nonexisting_resource(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.env_assist.assert_raise_library_error(
            lambda: ticket_command.create_with_set(
                self.env_assist.get_env(),
                [self.set("resource1", "resX"), self.set("resY", "resource2")],
                {"ticket": "ticket1"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id=res_id,
                    expected_types=[],
                    context_type="",
                    context_id="",
                )
                for res_id in ("resX", "resY")
            ]
        )

    def test_validation(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.env_assist.assert_raise_library_error(
            lambda: ticket_command.create_with_set(
                self.env_assist.get_env(),
                [
                    self.set(
                        "resource1",
                        "resource1-meta",
                        action="bad_action",
                        sequential="bad_sequential",
                        set_option="value",
                    ),
                    self.set(
                        "resource2-meta",
                        "resource4",
                        role="bad_role",
                        require_all="bad_require_all",
                    ),
                    self.set(),
                ],
                {
                    "ticket": "ticket_1",
                    "id": "resource3",
                    "loss-policy": "bad_policy",
                    "top-option": "value",
                },
            )
        )
        pcmk_bool = (
            "a pacemaker boolean value: '0', '1', 'false', 'n', 'no', 'off', "
            "'on', 'true', 'y', 'yes'"
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="resource1-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["set_option"],
                    allowed=["action", "require-all", "role", "sequential"],
                    option_type="set",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="action",
                    option_value="bad_action",
                    allowed_values=("start", "stop", "promote", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="sequential",
                    option_value="bad_sequential",
                    allowed_values=pcmk_bool,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="resource2-meta",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="require-all",
                    option_value="bad_require_all",
                    allowed_values=pcmk_bool,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value="bad_role",
                    allowed_values=(
                        "Stopped",
                        "Started",
                        "Promoted",
                        "Unpromoted",
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(reports.codes.EMPTY_RESOURCE_SET),
                fixture.error(
                    reports.codes.BOOTH_TICKET_NAME_INVALID,
                    ticket_name="ticket_1",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["top-option"],
                    allowed=["id", "loss-policy", "ticket"],
                    option_type=None,
                    allowed_patterns=[],
                ),
                fixture.error(reports.codes.ID_ALREADY_EXISTS, id="resource3"),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="loss-policy",
                    option_value="bad_policy",
                    allowed_values=("fence", "stop", "freeze", "demote"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_duplicate_constraint(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources=self.cib_resources,
            constraints="""
                <constraints>
                    <rsc_ticket id="ticket_set_r1r2r3" ticket="ticket1">
                        <resource_set id="ticket_set_r1r2r3_set">
                            <resource_ref id="resource1" />
                            <resource_ref id="resource2" />
                        </resource_set>
                        <resource_set id="ticket_set_r1r2r3_set-1">
                            <resource_ref id="resource3" />
                            <resource_ref id="resource4" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: ticket_command.create_with_set(
                self.env_assist.get_env(),
                [
                    self.set("resource1", "resource2"),
                    self.set("resource3", "resource4"),
                ],
                {"ticket": "ticket1"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    force_code=reports.codes.FORCE,
                    constraint_ids=["ticket_set_r1r2r3"],
                )
            ]
        )

    def test_duplicate_constraint_forced(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml",
            resources=self.cib_resources,
            constraints="""
                <constraints>
                    <rsc_ticket id="ticket_set_r1r2r3" ticket="ticket1">
                        <resource_set id="ticket_set_r1r2r3_set">
                            <resource_ref id="resource1" />
                            <resource_ref id="resource2" />
                        </resource_set>
                        <resource_set id="ticket_set_r1r2r3_set-1">
                            <resource_ref id="resource3" />
                            <resource_ref id="resource4" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """,
        )
        self.config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket_set_r1r2r3" ticket="ticket1">
                        <resource_set id="ticket_set_r1r2r3_set">
                            <resource_ref id="resource1" />
                            <resource_ref id="resource2" />
                        </resource_set>
                        <resource_set id="ticket_set_r1r2r3_set-1">
                            <resource_ref id="resource3" />
                            <resource_ref id="resource4" />
                        </resource_set>
                    </rsc_ticket>
                    <rsc_ticket id="ticket_set_r1r2r3-1" ticket="ticket1">
                        <resource_set id="ticket_set_r1r2r3-1_set">
                            <resource_ref id="resource1" />
                            <resource_ref id="resource2" />
                        </resource_set>
                        <resource_set id="ticket_set_r1r2r3-1_set-1">
                            <resource_ref id="resource3" />
                            <resource_ref id="resource4" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """
        )
        ticket_command.create_with_set(
            self.env_assist.get_env(),
            [
                self.set("resource1", "resource2"),
                self.set("resource3", "resource4"),
            ],
            {"ticket": "ticket1"},
            duplication_alowed=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    constraint_ids=["ticket_set_r1r2r3"],
                )
            ]
        )

    def test_resource_in_clone(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.env_assist.assert_raise_library_error(
            lambda: ticket_command.create_with_set(
                self.env_assist.get_env(),
                [
                    self.set("resource1", "resource-in-clone"),
                    self.set("resource-in-bundle", "resource2"),
                ],
                {"ticket": "ticket1"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="resource-in-clone",
                    parent_type="clone",
                    parent_id="C",
                ),
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id="resource-in-bundle",
                    parent_type="bundle",
                    parent_id="B",
                ),
            ]
        )

    def test_resource_in_clone_forced(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.7.xml", resources=self.cib_resources
        )
        self.config.env.push_cib(
            optional_in_conf="""
                <constraints>
                    <rsc_ticket id="ticket_set_r1rere" ticket="ticket1">
                        <resource_set id="ticket_set_r1rere_set">
                            <resource_ref id="resource1" />
                            <resource_ref id="resource-in-clone" />
                        </resource_set>
                        <resource_set id="ticket_set_r1rere_set-1">
                            <resource_ref id="resource-in-bundle" />
                            <resource_ref id="resource2" />
                        </resource_set>
                    </rsc_ticket>
                </constraints>
            """
        )
        ticket_command.create_with_set(
            self.env_assist.get_env(),
            [
                self.set("resource1", "resource-in-clone"),
                self.set("resource-in-bundle", "resource2"),
            ],
            {"ticket": "ticket1"},
            resource_in_clone_alowed=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="resource-in-clone",
                    parent_type="clone",
                    parent_id="C",
                ),
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id="resource-in-bundle",
                    parent_type="bundle",
                    parent_id="B",
                ),
            ]
        )


@patch_commands("get_constraints", mock.Mock())
class RemoveTest(TestCase):
    @patch_commands("ticket.remove_plain", mock.Mock(return_value=1))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=0)
    )
    def test_successfully_remove_plain(self):
        self.assertTrue(ticket_command.remove(mock.MagicMock(), "T", "R"))

    @patch_commands("ticket.remove_plain", mock.Mock(return_value=0))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=1)
    )
    def test_successfully_remove_with_resource_set(self):
        self.assertTrue(ticket_command.remove(mock.MagicMock(), "T", "R"))

    @patch_commands("ticket.remove_plain", mock.Mock(return_value=0))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=0)
    )
    def test_raises_library_error_when_no_matching_constraint_found(self):
        self.assertFalse(ticket_command.remove(mock.MagicMock(), "T", "R"))



from unittest import TestCase

from lxml import etree

from pcs.lib.cib.resource import common
from pcs.lib.cib.tools import IdProvider

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.xml import etree_to_str

fixture_cib = etree.fromstring(
    """
    <cib>
        <configuration>
            <resources>
                <primitive id="A" />
                <clone id="B-clone">
                    <primitive id="B" />
                </clone>
                <master id="C-master">
                    <primitive id="C" />
                </master>
                <group id="D">
                    <primitive id="D1" />
                    <primitive id="D2" />
                </group>
                <clone id="E-clone">
                    <group id="E">
                        <primitive id="E1" />
                        <primitive id="E2" />
                    </group>
                </clone>
                <master id="F-master">
                    <group id="F">
                        <primitive id="F1" />
                        <primitive id="F2" />
                    </group>
                </master>
                <bundle id="G-bundle" />
                <bundle id="H-bundle">
                    <primitive id="H" />
                </bundle>
                <group id="I">
                    <primitive id="I1" />
                </group>
                <clone id="J-clone">
                    <group id="J">
                        <primitive id="J1" />
                    </group>
                </clone>
                <master id="K-master">
                    <group id="K">
                        <primitive id="K1" />
                    </group>
                </master>
            </resources>
        </configuration>
    </cib>
    """
)


class FindOneOrMoreResources(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
                <primitive id="R1x" />
                <primitive id="R2x" />
            </resources>
        """
        )

        def searcher(resource_element):
            return [
                resource_element.getparent().find(
                    ".//*[@id='{0}x']".format(resource_element.get("id"))
                )
            ]

        self.additional_search = searcher

    def test_one_existing(self):
        resource, report_list = common.find_one_resource(self.cib, "R1")
        self.assertEqual("R1", resource.attrib.get("id"))
        assert_report_item_list_equal(report_list, [])

    def test_one_nonexistent(self):
        resource, report_list = common.find_one_resource(self.cib, "R-missing")
        self.assertIsNone(resource)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.report_not_found("R-missing", context_type="resources"),
            ],
        )

    def test_more_existing(self):
        resource_list, report_list = common.find_resources(
            self.cib, ["R1", "R2"]
        )
        self.assertEqual(
            ["R1", "R2"],
            [resource.attrib.get("id") for resource in resource_list],
        )
        assert_report_item_list_equal(report_list, [])

    def test_more_some_missing(self):
        resource_list, report_list = common.find_resources(
            self.cib, ["R1", "R2", "RY1", "RY2"]
        )
        self.assertEqual(
            ["R1", "R2"],
            [resource.attrib.get("id") for resource in resource_list],
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.report_not_found("RY1", context_type="resources"),
                fixture.report_not_found("RY2", context_type="resources"),
            ],
        )


class FindResourcesMixin:
    _iterable_type = list

    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            self._iterable_type(output_resource_ids),
            self._iterable_type(
                [
                    element.get("id", "")
                    for element in self._tested_fn(
                        fixture_cib.find(
                            './/*[@id="{0}"]'.format(input_resource_id)
                        )
                    )
                ]
            ),
        )

    def test_group(self):
        self.assert_find_resources("D", ["D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", [])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H"])

    def test_primitive(self):
        raise NotImplementedError()

    def test_primitive_in_clone(self):
        raise NotImplementedError()

    def test_primitive_in_master(self):
        raise NotImplementedError()

    def test_primitive_in_group(self):
        raise NotImplementedError()

    def test_primitive_in_bundle(self):
        raise NotImplementedError()

    def test_cloned_group(self):
        raise NotImplementedError()

    def test_mastered_group(self):
        raise NotImplementedError()


class FindPrimitives(TestCase, FindResourcesMixin):
    _tested_fn = staticmethod(common.find_primitives)

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E1", "E2"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F1", "F2"])


class GetAllInnerResources(TestCase, FindResourcesMixin):
    _iterable_type = set
    _tested_fn = staticmethod(common.get_all_inner_resources)

    def test_primitive(self):
        self.assert_find_resources("A", set())

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", set())

    def test_primitive_in_master(self):
        self.assert_find_resources("C", set())

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", set())
        self.assert_find_resources("D2", set())
        self.assert_find_resources("E1", set())
        self.assert_find_resources("E2", set())
        self.assert_find_resources("F1", set())
        self.assert_find_resources("F2", set())

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", set())

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", {"E", "E1", "E2"})

    def test_mastered_group(self):
        self.assert_find_resources("F-master", {"F", "F1", "F2"})


class GetInnerResources(TestCase, FindResourcesMixin):
    _tested_fn = staticmethod(common.get_inner_resources)

    def test_primitive(self):
        self.assert_find_resources("A", [])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", [])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", [])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", [])
        self.assert_find_resources("D2", [])
        self.assert_find_resources("E1", [])
        self.assert_find_resources("E2", [])
        self.assert_find_resources("F1", [])
        self.assert_find_resources("F2", [])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", [])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E"])


class IsWrapperResource(TestCase):
    def assert_is_wrapper(self, res_id, is_wrapper):
        self.assertEqual(
            is_wrapper,
            common.is_wrapper_resource(
                fixture_cib.find('.//*[@id="{0}"]'.format(res_id))
            ),
        )

    def test_primitive(self):
        self.assert_is_wrapper("A", False)

    def test_primitive_in_clone(self):
        self.assert_is_wrapper("B", False)

    def test_primitive_in_master(self):
        self.assert_is_wrapper("C", False)

    def test_primitive_in_group(self):
        self.assert_is_wrapper("D1", False)
        self.assert_is_wrapper("D2", False)
        self.assert_is_wrapper("E1", False)
        self.assert_is_wrapper("E2", False)
        self.assert_is_wrapper("F1", False)
        self.assert_is_wrapper("F2", False)

    def test_primitive_in_bundle(self):
        self.assert_is_wrapper("H", False)

    def test_cloned_group(self):
        self.assert_is_wrapper("E-clone", True)

    def test_mastered_group(self):
        self.assert_is_wrapper("F-master", True)

    def test_group(self):
        self.assert_is_wrapper("D", True)

    def test_group_in_clone(self):
        self.assert_is_wrapper("E", True)

    def test_group_in_master(self):
        self.assert_is_wrapper("F", True)

    def test_cloned_primitive(self):
        self.assert_is_wrapper("B-clone", True)

    def test_mastered_primitive(self):
        self.assert_is_wrapper("C-master", True)

    def test_bundle_empty(self):
        self.assert_is_wrapper("G-bundle", True)

    def test_bundle_with_primitive(self):
        self.assert_is_wrapper("H-bundle", True)


class GetParentResource(TestCase):
    def assert_parent_resource(self, input_resource_id, output_resource_id):
        res_el = common.get_parent_resource(
            fixture_cib.find('.//*[@id="{0}"]'.format(input_resource_id))
        )
        self.assertEqual(
            output_resource_id, res_el.get("id") if res_el is not None else None
        )

    def test_primitive(self):
        self.assert_parent_resource("A", None)

    def test_primitive_in_clone(self):
        self.assert_parent_resource("B", "B-clone")

    def test_primitive_in_master(self):
        self.assert_parent_resource("C", "C-master")

    def test_primitive_in_group(self):
        self.assert_parent_resource("D1", "D")
        self.assert_parent_resource("D2", "D")
        self.assert_parent_resource("E1", "E")
        self.assert_parent_resource("E2", "E")
        self.assert_parent_resource("F1", "F")
        self.assert_parent_resource("F2", "F")

    def test_primitive_in_bundle(self):
        self.assert_parent_resource("H", "H-bundle")

    def test_cloned_group(self):
        self.assert_parent_resource("E-clone", None)

    def test_mastered_group(self):
        self.assert_parent_resource("F-master", None)

    def test_group(self):
        self.assert_parent_resource("D", None)

    def test_group_in_clone(self):
        self.assert_parent_resource("E", "E-clone")

    def test_group_in_master(self):
        self.assert_parent_resource("F", "F-master")

    def test_cloned_primitive(self):
        self.assert_parent_resource("B-clone", None)

    def test_mastered_primitive(self):
        self.assert_parent_resource("C-master", None)

    def test_bundle_empty(self):
        self.assert_parent_resource("G-bundle", None)

    def test_bundle_with_primitive(self):
        self.assert_parent_resource("H-bundle", None)


class FindResourcesToEnable(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_enable(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B", "B-clone"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C", "C-master"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H", "H-bundle"])

    def test_group(self):
        self.assert_find_resources("D", ["D"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E", "E-clone"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F", "F-master"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B-clone", "B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E-clone", "E"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C-master", "C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F-master", "F"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class Enable(TestCase):
    @staticmethod
    def assert_enabled(pre, post):
        resource = etree.fromstring(pre)
        common.enable(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_disabled(self):
        self.assert_enabled(
            """
                <resource>
                    <meta_attributes>
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes />
                </resource>
            """,
        )

    def test_enabled(self):
        self.assert_enabled(
            """
                <resource>
                </resource>
            """,
            """
                <resource>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_enabled(
            """
                <resource>
                    <meta_attributes id="meta1">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                    <meta_attributes id="meta2">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes id="meta1" />
                    <meta_attributes id="meta2">
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </resource>
            """,
        )


class Disable(TestCase):
    @staticmethod
    def assert_disabled(pre, post):
        resource = etree.fromstring(pre)
        common.disable(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_disabled(self):
        xml = """
            <resource id="R">
                <meta_attributes id="R-meta_attributes">
                    <nvpair id="R-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </resource>
        """
        self.assert_disabled(xml, xml)

    def test_enabled(self):
        self.assert_disabled(
            """
                <resource id="R">
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_disabled(
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-target-role"
                            name="target-role" value="Stopped" />
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
        )


class IsDisabled(TestCase):
    def test_success(self):
        # Using non-existing tag "res" as the function doesn't care whether it's
        # a primitive, group or clone
        resources = etree.fromstring(
            """
            <resources>
                <res id="R1" />
                <res id="R2">
                    <meta_attributes>
                        <nvpair name="target-role" value="Stopped" />
                    </meta_attributes>
                </res>
                <res id="R3">
                    <meta_attributes>
                        <nvpair name="target-role" value="stopped" />
                    </meta_attributes>
                </res>
                <res id="R4">
                    <meta_attributes>
                        <nvpair name="target-role" value="something" />
                    </meta_attributes>
                </res>
            </resources>
            """
        )
        res_disabled = (
            ("R1", False),
            ("R2", True),
            ("R3", True),
            ("R4", False),
        )
        for res_name, expected_disabled in res_disabled:
            with self.subTest(
                resource=res_name, expected_disabled=expected_disabled
            ):
                self.assertEqual(
                    common.is_disabled(
                        resources.find(f".//*[@id='{res_name}']")
                    ),
                    expected_disabled,
                )


class FindResourcesToManage(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_manage(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B", "B-clone"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C", "C-master"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1", "D"])
        self.assert_find_resources("D2", ["D2", "D"])
        self.assert_find_resources("E1", ["E1", "E-clone", "E"])
        self.assert_find_resources("E2", ["E2", "E-clone", "E"])
        self.assert_find_resources("F1", ["F1", "F-master", "F"])
        self.assert_find_resources("F2", ["F2", "F-master", "F"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H", "H-bundle"])

    def test_group(self):
        self.assert_find_resources("D", ["D", "D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E", "E-clone", "E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F", "F-master", "F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B-clone", "B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E-clone", "E", "E1", "E2"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C-master", "C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F-master", "F", "F1", "F2"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class FindResourcesToUnmanage(TestCase):
    def assert_find_resources(self, input_resource_id, output_resource_ids):
        self.assertEqual(
            output_resource_ids,
            [
                element.get("id", "")
                for element in common.find_resources_to_unmanage(
                    fixture_cib.find(
                        './/*[@id="{0}"]'.format(input_resource_id)
                    )
                )
            ],
        )

    def test_primitive(self):
        self.assert_find_resources("A", ["A"])

    def test_primitive_in_clone(self):
        self.assert_find_resources("B", ["B"])

    def test_primitive_in_master(self):
        self.assert_find_resources("C", ["C"])

    def test_primitive_in_group(self):
        self.assert_find_resources("D1", ["D1"])
        self.assert_find_resources("D2", ["D2"])
        self.assert_find_resources("E1", ["E1"])
        self.assert_find_resources("E2", ["E2"])
        self.assert_find_resources("F1", ["F1"])
        self.assert_find_resources("F2", ["F2"])

    def test_primitive_in_bundle(self):
        self.assert_find_resources("H", ["H"])

    def test_group(self):
        self.assert_find_resources("D", ["D1", "D2"])

    def test_group_in_clone(self):
        self.assert_find_resources("E", ["E1", "E2"])

    def test_group_in_master(self):
        self.assert_find_resources("F", ["F1", "F2"])

    def test_cloned_primitive(self):
        self.assert_find_resources("B-clone", ["B"])

    def test_cloned_group(self):
        self.assert_find_resources("E-clone", ["E1", "E2"])

    def test_mastered_primitive(self):
        self.assert_find_resources("C-master", ["C"])

    def test_mastered_group(self):
        self.assert_find_resources("F-master", ["F1", "F2"])

    def test_bundle_empty(self):
        self.assert_find_resources("G-bundle", ["G-bundle"])

    def test_bundle_with_primitive(self):
        self.assert_find_resources("H-bundle", ["H-bundle", "H"])


class Manage(TestCase):
    @staticmethod
    def assert_managed(pre, post):
        resource = etree.fromstring(pre)
        common.manage(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_unmanaged(self):
        self.assert_managed(
            """
                <resource>
                    <meta_attributes>
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes />
                </resource>
            """,
        )

    def test_managed(self):
        self.assert_managed(
            """
                <resource>
                </resource>
            """,
            """
                <resource>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_managed(
            """
                <resource>
                    <meta_attributes id="meta1">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                    <meta_attributes id="meta2">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
            """
                <resource>
                    <meta_attributes id="meta1" />
                    <meta_attributes id="meta2">
                        <nvpair name="is-managed" value="something" />
                    </meta_attributes>
                </resource>
            """,
        )


class Unmanage(TestCase):
    @staticmethod
    def assert_unmanaged(pre, post):
        resource = etree.fromstring(pre)
        common.unmanage(resource, IdProvider(resource))
        assert_xml_equal(post, etree_to_str(resource))

    def test_unmanaged(self):
        xml = """
            <resource id="R">
                <meta_attributes id="R-meta_attributes">
                    <nvpair id="R-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </resource>
        """
        self.assert_unmanaged(xml, xml)

    def test_managed(self):
        self.assert_unmanaged(
            """
                <resource id="R">
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </resource>
            """,
        )

    def test_only_first_meta(self):
        # this captures the current behavior
        # once pcs supports more instance and meta attributes for each resource,
        # this test should be reconsidered
        self.assert_unmanaged(
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
            """
                <resource id="R">
                    <meta_attributes id="R-meta_attributes">
                        <nvpair id="R-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <meta_attributes id="R-meta_attributes-2">
                    </meta_attributes>
                </resource>
            """,
        )



from pcs import (
    alert,
    usage,
)
from pcs.cli.alert import command as alert_command
from pcs.cli.common.errors import raise_command_replaced
from pcs.cli.common.routing import create_router

alert_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.alert(argv)),
        "create": alert.alert_add,
        "update": alert.alert_update,
        "delete": alert.alert_remove,
        "remove": alert.alert_remove,
        "config": alert_command.alert_config,
        "show": lambda lib, argv, modifiers: raise_command_replaced(
            ["pcs alert config"], pcs_version="0.12"
        ),
        "recipient": create_router(
            {
                "help": lambda lib, argv, modifiers: print(
                    usage.alert(["recipient"])
                ),
                "add": alert.recipient_add,
                "update": alert.recipient_update,
                "delete": alert.recipient_remove,
                "remove": alert.recipient_remove,
            },
            ["alert", "recipient"],
        ),
        "get_all_alerts": alert.print_alerts_in_json,
    },
    ["alert"],
    default_cmd="config",
)



import json
from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import lines_to_str
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    Argv,
    InputModifiers,
)
from pcs.common.interface.dto import to_dict

from .output import config_dto_to_cmd, config_dto_to_lines


def alert_config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported("-f", output_format_supported=True)
    output_format = modifiers.get_output_format()
    if argv:
        raise CmdLineInputError

    config_dto = lib.alert.get_config_dto()

    if output_format == OUTPUT_FORMAT_VALUE_JSON:
        print(json.dumps(to_dict(config_dto), indent=2))
        return

    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        result_cmd = config_dto_to_cmd(config_dto)
        if result_cmd:
            print(";\n".join(result_cmd))
        return

    result_text = lines_to_str(config_dto_to_lines(config_dto))
    if result_text:
        print(result_text)



from pcs.common.const import (
    PcmkAction,
    PcmkRoleType,
)
from pcs.common.pacemaker.constraint import (
    CibConstraintColocationAttributesDto,
    CibConstraintColocationDto,
    CibConstraintColocationSetDto,
    CibConstraintLocationAttributesDto,
    CibConstraintLocationDto,
    CibConstraintOrderAttributesDto,
    CibConstraintOrderDto,
    CibConstraintOrderSetDto,
    CibConstraintsDto,
    CibConstraintTicketAttributesDto,
    CibConstraintTicketDto,
    CibConstraintTicketSetDto,
    CibResourceSetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.pacemaker.types import (
    CibResourceDiscovery,
    CibResourceSetOrderType,
    CibTicketLossPolicy,
)
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)
from pcs.lib.cib.rule.in_effect import RuleInEffectEval


def get_all_constraints(
    rule_eval: RuleInEffectEval,
    include_expired: bool = True,
) -> CibConstraintsDto:
    location = [
        CibConstraintLocationDto(
            resource_id="R7",
            resource_pattern=None,
            role=None,
            attributes=CibConstraintLocationAttributesDto(
                constraint_id="location-R7-non-existing-node--10000",
                score="-10000",
                node="non-existing-node",
                rules=[],
                lifetime=[],
                resource_discovery=None,
            ),
        ),
        CibConstraintLocationDto(
            resource_id="R7",
            resource_pattern=None,
            role=None,
            attributes=CibConstraintLocationAttributesDto(
                constraint_id="location-R7-another-one--INFINITY",
                score="-INFINITY",
                node="another-one",
                rules=[],
                lifetime=[],
                resource_discovery=None,
            ),
        ),
        CibConstraintLocationDto(
            resource_id="R7",
            resource_pattern=None,
            role=None,
            attributes=CibConstraintLocationAttributesDto(
                constraint_id="location-R7-localhost-INFINITY",
                score="INFINITY",
                node="localhost",
                rules=[],
                lifetime=[],
                resource_discovery=CibResourceDiscovery.ALWAYS,
            ),
        ),
        CibConstraintLocationDto(
            resource_id="G2",
            resource_pattern=None,
            role=None,
            attributes=CibConstraintLocationAttributesDto(
                constraint_id="location-G2-localhost-INFINITY",
                score="INFINITY",
                node="localhost",
                rules=[],
                lifetime=[],
                resource_discovery=None,
            ),
        ),
        CibConstraintLocationDto(
            resource_id=None,
            resource_pattern="R*",
            role=None,
            attributes=CibConstraintLocationAttributesDto(
                constraint_id="location-R-localhost-INFINITY",
                score="INFINITY",
                node="localhost",
                rules=[],
                lifetime=[],
                resource_discovery=None,
            ),
        ),
    ]
    location.extend(
        loc_const
        for loc_const in [
            CibConstraintLocationDto(
                resource_id="B2",
                resource_pattern=None,
                role=None,
                attributes=CibConstraintLocationAttributesDto(
                    constraint_id="loc_constr_with_expired_rule",
                    score=None,
                    node=None,
                    rules=[
                        CibRuleExpressionDto(
                            id="loc_constr_with_expired_rule-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status(
                                "loc_constr_with_expired_rule-rule"
                            ),
                            options={
                                "boolean-op": "and",
                                "score": "500",
                            },
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="loc_constr_with_expired_rule-rule-expr",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=rule_eval.get_rule_status(
                                        "loc_constr_with_expired_rule-rule-expr"
                                    ),
                                    options={
                                        "operation": "lt",
                                        "end": "2000-01-01",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date lt 2000-01-01",
                                )
                            ],
                            as_string="date lt 2000-01-01",
                        )
                    ],
                    lifetime=[],
                    resource_discovery=None,
                ),
            ),
            CibConstraintLocationDto(
                resource_id="R6-clone",
                resource_pattern=None,
                role=None,
                attributes=CibConstraintLocationAttributesDto(
                    constraint_id="loc_constr_with_not_expired_rule",
                    score=None,
                    node=None,
                    rules=[
                        CibRuleExpressionDto(
                            id="loc_constr_with_not_expired_rule-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status(
                                "loc_constr_with_not_expired_rule-rule"
                            ),
                            options={
                                "boolean-op": "and",
                                "role": "Unpromoted",
                                "score": "500",
                            },
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="loc_constr_with_not_expired_rule-rule-expr",
                                    type=CibRuleExpressionType.EXPRESSION,
                                    in_effect=rule_eval.get_rule_status(
                                        "loc_constr_with_not_expired_rule-rule-expr"
                                    ),
                                    options={
                                        "operation": "eq",
                                        "attribute": "#uname",
                                        "value": "node1",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="#uname eq node1",
                                ),
                                CibRuleExpressionDto(
                                    id="loc_constr_with_not_expired_rule-rule-expr-1",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=rule_eval.get_rule_status(
                                        "loc_constr_with_not_expired_rule-rule-expr-1"
                                    ),
                                    options={
                                        "operation": "gt",
                                        "start": "2000-01-01",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date gt 2000-01-01",
                                ),
                            ],
                            as_string="#uname eq node1 and date gt 2000-01-01",
                        ),
                    ],
                    lifetime=[],
                    resource_discovery=None,
                ),
            ),
            CibConstraintLocationDto(
                resource_id="R6-clone",
                resource_pattern=None,
                role=None,
                attributes=CibConstraintLocationAttributesDto(
                    constraint_id="loc_constr_with_not_expired_rule-1",
                    score=None,
                    node=None,
                    rules=[
                        CibRuleExpressionDto(
                            id="loc_constr_with_not_expired_rule-1-rule",
                            type=CibRuleExpressionType.RULE,
                            in_effect=rule_eval.get_rule_status(
                                "loc_constr_with_not_expired_rule-1-rule"
                            ),
                            options={
                                "boolean-op": "and",
                                "role": "Promoted",
                                "score-attribute": "test-attr",
                            },
                            date_spec=None,
                            duration=None,
                            expressions=[
                                CibRuleExpressionDto(
                                    id="loc_constr_with_not_expired_rule-1-rule-expr",
                                    type=CibRuleExpressionType.DATE_EXPRESSION,
                                    in_effect=rule_eval.get_rule_status(
                                        "loc_constr_with_not_expired_rule-1-rule-expr"
                                    ),
                                    options={
                                        "operation": "gt",
                                        "start": "2010-12-31",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="date gt 2010-12-31",
                                ),
                                CibRuleExpressionDto(
                                    id="loc_constr_with_not_expired_rule-1-rule-expr-1",
                                    type=CibRuleExpressionType.EXPRESSION,
                                    in_effect=rule_eval.get_rule_status(
                                        "loc_constr_with_not_expired_rule-1-rule-expr-1"
                                    ),
                                    options={
                                        "operation": "eq",
                                        "attribute": "#uname",
                                        "value": "node1",
                                    },
                                    date_spec=None,
                                    duration=None,
                                    expressions=[],
                                    as_string="#uname eq node1",
                                ),
                            ],
                            as_string="date gt 2010-12-31 and #uname eq node1",
                        ),
                    ],
                    lifetime=[],
                    resource_discovery=None,
                ),
            ),
        ]
        if include_expired
        or not any(
            rule.in_effect == CibRuleInEffectStatus.EXPIRED
            for rule in loc_const.attributes.rules
        )
    )

    return CibConstraintsDto(
        location=location,
        location_set=[],
        colocation=[
            CibConstraintColocationDto(
                resource_id="G1-clone",
                with_resource_id="R6-clone",
                node_attribute=None,
                resource_role=PcmkRoleType("Promoted"),
                with_resource_role=PcmkRoleType("Stopped"),
                resource_instance=None,
                with_resource_instance=None,
                attributes=CibConstraintColocationAttributesDto(
                    constraint_id="colocation-G1-clone-R6-clone--100",
                    score="-100",
                    influence=None,
                    lifetime=[],
                ),
            )
        ],
        colocation_set=[
            CibConstraintColocationSetDto(
                resource_sets=[
                    CibResourceSetDto(
                        set_id="colocation_set_R7G2B2_set",
                        sequential=None,
                        require_all=None,
                        ordering=None,
                        action=None,
                        role=PcmkRoleType("Started"),
                        score=None,
                        kind=None,
                        resources_ids=["R7", "G2"],
                    ),
                    CibResourceSetDto(
                        set_id="colocation_set_R7G2B2_set-1",
                        sequential=False,
                        require_all=None,
                        ordering=None,
                        action=None,
                        role=None,
                        score=None,
                        kind=None,
                        resources_ids=["B2", "R6-clone"],
                    ),
                ],
                attributes=CibConstraintColocationAttributesDto(
                    constraint_id="colocation_set_R7G2B2",
                    score="-1",
                    influence=None,
                    lifetime=[],
                ),
            )
        ],
        order=[
            CibConstraintOrderDto(
                first_resource_id="R7",
                then_resource_id="G2",
                first_action=PcmkAction("stop"),
                then_action=PcmkAction("stop"),
                first_resource_instance=None,
                then_resource_instance=None,
                attributes=CibConstraintOrderAttributesDto(
                    constraint_id="order-R7-G2-mandatory",
                    symmetrical=False,
                    require_all=False,
                    score="-123",
                    kind=None,
                ),
            ),
            CibConstraintOrderDto(
                first_resource_id="G2",
                then_resource_id="B2",
                first_action=PcmkAction("start"),
                then_action=PcmkAction("start"),
                first_resource_instance=None,
                then_resource_instance=None,
                attributes=CibConstraintOrderAttributesDto(
                    constraint_id="order-G2-B2-Optional",
                    symmetrical=None,
                    require_all=None,
                    score=None,
                    kind=CibResourceSetOrderType.OPTIONAL,
                ),
            ),
        ],
        order_set=[
            CibConstraintOrderSetDto(
                resource_sets=[
                    CibResourceSetDto(
                        set_id="order_set_B2R6-cloneSe_set",
                        sequential=None,
                        require_all=False,
                        ordering=None,
                        action=PcmkAction("stop"),
                        role=None,
                        score=None,
                        kind=None,
                        resources_ids=["B2", "R6-clone"],
                    ),
                    CibResourceSetDto(
                        set_id="order_set_B2R6-cloneSe_set-1",
                        sequential=False,
                        require_all=None,
                        ordering=None,
                        action=PcmkAction("promote"),
                        role=None,
                        score=None,
                        kind=None,
                        resources_ids=["G1-clone"],
                    ),
                ],
                attributes=CibConstraintOrderAttributesDto(
                    constraint_id="order_set_B2R6-cloneSe",
                    symmetrical=None,
                    require_all=None,
                    score=None,
                    kind=CibResourceSetOrderType.OPTIONAL,
                ),
            )
        ],
        ticket=[
            CibConstraintTicketDto(
                resource_id="G1-clone",
                role=PcmkRoleType("Promoted"),
                attributes=CibConstraintTicketAttributesDto(
                    constraint_id="ticket-custom-ticket1-G1-clone-Promoted",
                    ticket="custom-ticket1",
                    loss_policy=CibTicketLossPolicy.DEMOTE,
                ),
            )
        ],
        ticket_set=[
            CibConstraintTicketSetDto(
                resource_sets=[
                    CibResourceSetDto(
                        set_id="ticket_set_R7B2G2_set",
                        sequential=None,
                        require_all=None,
                        ordering=None,
                        action=None,
                        role=PcmkRoleType("Stopped"),
                        score=None,
                        kind=None,
                        resources_ids=["R7", "B2", "G2"],
                    )
                ],
                attributes=CibConstraintTicketAttributesDto(
                    constraint_id="ticket_set_R7B2G2",
                    ticket="ticket2",
                    loss_policy=None,
                ),
            )
        ],
    )



# pylint: disable=too-many-lines
from typing import (
    Optional,
    Sequence,
    Union,
)
from unittest import TestCase

from pcs.common.const import (
    PCMK_ROLE_STOPPED,
    PCMK_STATUS_ROLE_STARTED,
    PCMK_STATUS_ROLE_STOPPED,
    PCMK_STATUS_ROLES_PENDING,
    PcmkRoleType,
    PcmkStatusRoleType,
)
from pcs.common.resource_status import (
    InstancesQuantifierUnsupportedException,
    MembersQuantifierUnsupportedException,
    MoreChildrenQuantifierType,
    ResourceNonExistentException,
    ResourceNotInGroupException,
    ResourcesStatusFacade,
    ResourceState,
    ResourceType,
    ResourceUnexpectedTypeException,
    _filter_clone_orphans,
)
from pcs.common.status_dto import (
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)
from pcs.common.types import StringSequence


def fixture_primitive_dto(  # noqa: PLR0913
    resource_id: str,
    instance_id: Optional[str],
    *,
    resource_agent: str = "ocf:pacemaker:Dummy",
    role: PcmkStatusRoleType = PCMK_STATUS_ROLE_STARTED,
    target_role: Optional[PcmkRoleType] = None,
    active: bool = True,
    orphaned: bool = False,
    blocked: bool = False,
    maintenance: bool = False,
    failed: bool = False,
    failure_ignored: bool = False,
    managed: bool = True,
    node_names: StringSequence = ("node1",),
    pending: Optional[str] = None,
    locked_to: Optional[str] = None,
) -> PrimitiveStatusDto:
    # pylint: disable=too-many-arguments
    return PrimitiveStatusDto(
        resource_id,
        instance_id,
        resource_agent,
        role,
        target_role,
        active,
        orphaned,
        blocked,
        maintenance,
        None,
        failed,
        managed,
        failure_ignored,
        list(node_names),
        pending,
        locked_to,
    )


def fixture_group_dto(
    resource_id: str,
    instance_id: Optional[str],
    maintenance: bool = False,
    managed: bool = True,
    disabled: bool = False,
    members: Sequence[PrimitiveStatusDto] = (),
) -> GroupStatusDto:
    return GroupStatusDto(
        resource_id,
        instance_id,
        maintenance,
        None,
        managed,
        disabled,
        list(members),
    )


def fixture_clone_dto(  # noqa: PLR0913
    resource_id: str,
    *,
    multi_state: bool = False,
    unique: bool = False,
    maintenance: bool = False,
    managed: bool = True,
    disabled: bool = False,
    failed: bool = False,
    failure_ignored: bool = False,
    instances: Union[
        Sequence[PrimitiveStatusDto], Sequence[GroupStatusDto]
    ] = (),
) -> CloneStatusDto:
    # pylint: disable=too-many-arguments
    return CloneStatusDto(
        resource_id,
        multi_state,
        unique,
        maintenance,
        None,
        managed,
        disabled,
        failed,
        failure_ignored,
        None,
        instances,
    )


class TestFilterCloneOrphans(TestCase):
    # pylint: disable=protected-access
    def test_primitives(self):
        resource_list = [
            fixture_primitive_dto("primitive", None, node_names=["node1"]),
            fixture_primitive_dto("primitive", None, node_names=["node2"]),
            fixture_primitive_dto(
                "primitive", None, orphaned=True, node_names=["node3"]
            ),
        ]

        filtered_list = _filter_clone_orphans(resource_list)
        self.assertEqual(
            filtered_list,
            [
                fixture_primitive_dto("primitive", None, node_names=["node1"]),
                fixture_primitive_dto("primitive", None, node_names=["node2"]),
            ],
        )

    def test_groups(self):
        primitive_names = ["node1", "node2"]
        resource_list = [
            fixture_group_dto(
                "group",
                "0",
                members=[
                    fixture_primitive_dto(name, None)
                    for name in primitive_names
                ],
            ),
            fixture_group_dto(
                "group",
                "1",
                members=[
                    fixture_primitive_dto(
                        "primitive_1", None, orphaned=True, node_names=["node2"]
                    )
                    for name in primitive_names
                ],
            ),
        ]
        filtered_list = _filter_clone_orphans(resource_list)
        self.assertEqual(
            filtered_list,
            [
                fixture_group_dto(
                    "group",
                    "0",
                    members=[
                        fixture_primitive_dto(name, None)
                        for name in primitive_names
                    ],
                )
            ],
        )


class TestFacadeFromDto(TestCase):
    # pylint: disable=protected-access
    def test_primitive(self):
        primitive = fixture_primitive_dto("primitive", None)
        dto = ResourcesStatusDto([primitive])

        facade = ResourcesStatusFacade.from_resources_status_dto(dto)

        self.assertEqual(facade._resources, [primitive])
        self.assertEqual(facade._resource_map, {"primitive": [primitive]})
        self.assertEqual(facade._child_parent_map, {})
        self.assertEqual(facade._bundle_member_node_map, {})
        self.assertFalse(facade._bundle_defined)

    def test_group(self):
        primitive = fixture_primitive_dto("primitive", None)
        group = fixture_group_dto("group", None, members=[primitive])

        dto = ResourcesStatusDto([group])
        facade = ResourcesStatusFacade.from_resources_status_dto(dto)

        self.assertEqual(facade._resources, [group])
        self.assertEqual(
            facade._resource_map, {"group": [group], "primitive": [primitive]}
        )
        self.assertEqual(facade._child_parent_map, {"primitive": "group"})
        self.assertEqual(facade._bundle_member_node_map, {})
        self.assertFalse(facade._bundle_defined)

    def test_clone(self):
        primitive = fixture_primitive_dto("primitive", None)
        clone = fixture_clone_dto("clone", instances=[primitive])

        dto = ResourcesStatusDto([clone])
        facade = ResourcesStatusFacade.from_resources_status_dto(dto)

        self.assertEqual(facade._resources, [clone])
        self.assertEqual(
            facade._resource_map, {"clone": [clone], "primitive": [primitive]}
        )
        self.assertEqual(facade._child_parent_map, {"primitive": "clone"})
        self.assertEqual(facade._bundle_member_node_map, {})
        self.assertFalse(facade._bundle_defined)

    def test_cloned_group(self):
        primitive = fixture_primitive_dto("primitive", None)
        group = fixture_group_dto("group", None, members=[primitive])
        clone = fixture_clone_dto("clone", instances=[group])

        dto = ResourcesStatusDto([clone])
        facade = ResourcesStatusFacade.from_resources_status_dto(dto)

        self.assertEqual(facade._resources, [clone])
        self.assertEqual(
            facade._resource_map,
            {"clone": [clone], "group": [group], "primitive": [primitive]},
        )
        self.assertEqual(
            facade._child_parent_map, {"group": "clone", "primitive": "group"}
        )
        self.assertEqual(facade._bundle_member_node_map, {})
        self.assertFalse(facade._bundle_defined)

    def test_bundle(self):
        member = fixture_primitive_dto("member", None)
        container = fixture_primitive_dto("container", None)
        remote = fixture_primitive_dto("remote", None)
        bundle = BundleStatusDto(
            "bundle",
            "",
            "",
            False,
            False,
            None,
            True,
            False,
            [BundleReplicaStatusDto("0", member, remote, container, None)],
        )
        dto = ResourcesStatusDto([bundle])

        facade = ResourcesStatusFacade.from_resources_status_dto(dto)
        self.assertEqual(facade._resources, [bundle])
        self.assertEqual(
            facade._resource_map, {"bundle": [bundle], "member": [member]}
        )
        self.assertEqual(facade._child_parent_map, {"member": "bundle"})
        self.assertEqual(
            facade._bundle_member_node_map, {id(member): ["node1"]}
        )
        self.assertTrue(facade._bundle_defined)

    def test_combination(self):
        primitive = fixture_primitive_dto("primitive", None)
        primitive_member = fixture_primitive_dto("primitive_member", None)
        group = fixture_group_dto("group", None, members=[primitive_member])
        primitive_instance = fixture_primitive_dto("primitive_instance", None)
        clone = fixture_clone_dto("clone", instances=[primitive_instance])

        dto = ResourcesStatusDto([primitive, group, clone])
        facade = ResourcesStatusFacade.from_resources_status_dto(dto)

        self.assertEqual(facade._resources, [primitive, group, clone])
        self.assertEqual(
            facade._resource_map,
            {
                "primitive": [primitive],
                "group": [group],
                "primitive_member": [primitive_member],
                "clone": [clone],
                "primitive_instance": [primitive_instance],
            },
        )
        self.assertEqual(
            facade._child_parent_map,
            {"primitive_member": "group", "primitive_instance": "clone"},
        )
        self.assertEqual(facade._bundle_member_node_map, {})
        self.assertFalse(facade._bundle_defined)


def fixture_facade() -> ResourcesStatusFacade:
    stonith = fixture_primitive_dto(
        "stonith", None, resource_agent="stonith:fence_xvm"
    )
    primitive = fixture_primitive_dto("primitive", None)
    primitive_stopped = fixture_primitive_dto(
        "primitive_stopped",
        None,
        role=PCMK_STATUS_ROLE_STOPPED,
        active=False,
        node_names=[],
    )

    group_empty = fixture_group_dto("group_empty", None)
    group = fixture_group_dto(
        "group",
        None,
        members=[
            fixture_primitive_dto("group-primitive_1", None),
            fixture_primitive_dto("group-primitive_2", None),
        ],
    )

    clone_empty = fixture_clone_dto("clone_empty")
    clone = fixture_clone_dto(
        "clone",
        instances=[
            fixture_primitive_dto("clone-primitive", None, node_names=[node])
            for node in ["node1", "node2"]
        ],
    )

    clone_unique = fixture_clone_dto(
        "clone_unique",
        unique=True,
        # multiple replicas running on the same node
        instances=[
            fixture_primitive_dto(
                "clone_unique-primitive", str(i), node_names=[f"node{i % 2}"]
            )
            for i in range(4)
        ],
    )
    clone_promotable = fixture_clone_dto("clone_promotable", multi_state=True)

    cloned_group = fixture_clone_dto(
        "cloned_group",
        instances=[
            fixture_group_dto(
                "cloned_group-group",
                "0",
                members=[
                    fixture_primitive_dto(
                        "cloned_group-group-primitive_1", None
                    ),
                    fixture_primitive_dto(
                        "cloned_group-group-primitive_2",
                        None,
                        role=PCMK_STATUS_ROLE_STOPPED,
                        active=False,
                        node_names=[],
                    ),
                ],
            )
        ],
    )

    bundle_empty = BundleStatusDto(
        "bundle_empty",
        "podman",
        "",
        False,
        False,
        None,
        True,
        False,
        [],
    )

    bundle = BundleStatusDto(
        "bundle",
        "podman",
        "",
        True,
        False,
        None,
        True,
        False,
        [
            BundleReplicaStatusDto(
                str(i),
                member=fixture_primitive_dto(
                    "bundle-member", None, node_names=[f"bundle-remote-{i}"]
                ),
                remote=fixture_primitive_dto(
                    f"bundle-remote-{i}", None, node_names=[f"node{i}"]
                ),
                container=fixture_primitive_dto(
                    f"bundle-container-{i}", None, node_names=[f"node{i}"]
                ),
                ip_address=None,
            )
            for i in range(2)
        ],
    )

    return ResourcesStatusFacade(
        [
            stonith,
            primitive,
            primitive_stopped,
            group_empty,
            group,
            clone_empty,
            clone,
            clone_unique,
            clone_promotable,
            cloned_group,
            bundle_empty,
            bundle,
        ]
    )


class TestFacadeGetResource(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_one_instance_primitive(self):
        primitive = self.facade.get_resource_one_instance("primitive", None)
        self.assertEqual(primitive, fixture_primitive_dto("primitive", None))

    def test_one_instance_group(self):
        group = self.facade.get_resource_one_instance("group", None)
        self.assertEqual(
            group,
            fixture_group_dto(
                "group",
                None,
                members=[
                    fixture_primitive_dto("group-primitive_1", None),
                    fixture_primitive_dto("group-primitive_2", None),
                ],
            ),
        )

    def test_one_instance_clone(self):
        clone = self.facade.get_resource_one_instance("clone", None)
        self.assertEqual(
            clone,
            fixture_clone_dto(
                "clone",
                instances=[
                    fixture_primitive_dto(
                        "clone-primitive", None, node_names=[node]
                    )
                    for node in ["node1", "node2"]
                ],
            ),
        )

    def test_one_instance_clone_instance(self):
        clone_primitive = self.facade.get_resource_one_instance(
            "clone_unique-primitive", None
        )
        self.assertEqual(
            clone_primitive,
            fixture_primitive_dto(
                "clone_unique-primitive", "0", node_names=["node0"]
            ),
        )

    def test_one_instance_clone_instance_with_id(self):
        clone_primitive = self.facade.get_resource_one_instance(
            "clone_unique-primitive", "1"
        )
        self.assertEqual(
            clone_primitive,
            fixture_primitive_dto(
                "clone_unique-primitive", "1", node_names=["node1"]
            ),
        )

    def test_one_instance_bundle(self):
        bundle = self.facade.get_resource_one_instance("bundle", None)

        self.assertEqual(
            bundle,
            BundleStatusDto(
                "bundle",
                "podman",
                "",
                True,
                False,
                None,
                True,
                False,
                [
                    BundleReplicaStatusDto(
                        str(i),
                        member=fixture_primitive_dto(
                            "bundle-member",
                            None,
                            node_names=[f"bundle-remote-{i}"],
                        ),
                        remote=fixture_primitive_dto(
                            f"bundle-remote-{i}", None, node_names=[f"node{i}"]
                        ),
                        container=fixture_primitive_dto(
                            f"bundle-container-{i}",
                            None,
                            node_names=[f"node{i}"],
                        ),
                        ip_address=None,
                    )
                    for i in range(2)
                ],
            ),
        )

    def test_one_instance_bundle_member(self):
        member = self.facade.get_resource_one_instance("bundle-member", None)
        self.assertEqual(
            member,
            fixture_primitive_dto(
                "bundle-member", None, node_names=["bundle-remote-0"]
            ),
        )

    def test_one_instance_nonexistent(self):
        id_list = [
            ("nonexistent", None),
            ("primitive", "0"),
            ("clone-primitive", "42"),
            ("clone_unique-primitive", "42"),
            ("bundle-container", None),
        ]
        for resource_id, instance_id in id_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                resource = self.facade.get_resource_one_instance(
                    resource_id, instance_id
                )
                self.assertEqual(resource, None)

    def test_all_instances(self):
        clone = self.facade.get_resource_all_instances("clone_unique-primitive")
        self.assertEqual(
            clone,
            [
                fixture_primitive_dto(
                    "clone_unique-primitive",
                    str(i),
                    node_names=[f"node{i % 2}"],
                )
                for i in range(4)
            ],
        )

    def test_all_instances_nonexistent(self):
        for resource_id in ["nonexistent", "bundle-container"]:
            with self.subTest(value=resource_id):
                resource = self.facade.get_resource_all_instances(resource_id)
                self.assertEqual(resource, None)


class TestFacadeExists(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_exists(self):
        resource_id_list = [
            ("stonith", None),
            ("primitive", None),
            ("group", None),
            ("group-primitive_1", None),
            ("group-primitive_2", None),
            ("clone", None),
            ("clone-primitive", None),
            ("clone_unique", None),
            ("clone_unique-primitive", None),
            ("clone_unique-primitive", "0"),
            ("clone_unique-primitive", "1"),
            ("cloned_group", None),
            ("cloned_group-group", None),
            ("cloned_group-group", "0"),
            ("cloned_group-group-primitive_1", None),
            ("cloned_group-group-primitive_2", None),
            ("bundle", None),
            ("bundle-member", None),
        ]
        for resource_id, instance_id in resource_id_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                self.assertTrue(self.facade.exists(resource_id, instance_id))

    def test_does_not_exist(self):
        id_list = [
            ("nonexistent", None),
            ("primitive", "0"),
            ("clone-primitive", "42"),
            ("clone_unique-primitive", "42"),
            ("bundle-container", None),
        ]
        for resource_id, instance_id in id_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                self.assertFalse(self.facade.exists(resource_id, instance_id))


class TestFacadeResourceTypes(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_is_stonith(self):
        self.assertTrue(self.facade.is_stonith("stonith", None))
        self.assertFalse(self.facade.is_stonith("primitive", None))

    def test_is_stonith_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.is_stonith("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_get_type(self):
        resource_list = [
            ("primitive", None, ResourceType.PRIMITIVE),
            ("group", None, ResourceType.GROUP),
            ("clone", None, ResourceType.CLONE),
            ("cloned_group", None, ResourceType.CLONE),
            ("bundle", None, ResourceType.BUNDLE),
            ("group-primitive_1", None, ResourceType.PRIMITIVE),
            ("group-primitive_2", None, ResourceType.PRIMITIVE),
            ("clone_unique-primitive", None, ResourceType.PRIMITIVE),
            ("clone_unique-primitive", "0", ResourceType.PRIMITIVE),
            ("cloned_group-group", None, ResourceType.GROUP),
            ("cloned_group-group", "0", ResourceType.GROUP),
            ("cloned_group-group-primitive_1", None, ResourceType.PRIMITIVE),
            ("bundle-member", None, ResourceType.PRIMITIVE),
        ]
        for resource_id, instance_id, resource_type in resource_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                returned_type = self.facade.get_type(resource_id, instance_id)
                self.assertEqual(returned_type, resource_type)

    def test_get_type_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.get_type("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_is_promotable(self):
        self.assertTrue(self.facade.is_promotable("clone_promotable", None))
        self.assertFalse(self.facade.is_promotable("clone", None))

    def test_is_promotable_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.is_promotable("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_is_promotable_bad_type(self):
        resource_list = [
            ("primitive", ResourceType.PRIMITIVE),
            ("group", ResourceType.GROUP),
            ("bundle", ResourceType.BUNDLE),
        ]
        for resource_id, resource_type in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceUnexpectedTypeException) as cm:
                    self.facade.is_promotable(resource_id, None)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, None)
                self.assertEqual(cm.exception.resource_type, resource_type)
                self.assertEqual(
                    cm.exception.expected_types, [ResourceType.CLONE]
                )

    def test_is_unique_clone(self):
        self.assertTrue(self.facade.is_unique("clone_unique", None))
        self.assertFalse(self.facade.is_unique("clone", None))

    def test_is_unique_bundle(self):
        self.assertTrue(self.facade.is_unique("bundle", None))
        self.assertFalse(self.facade.is_unique("bundle_empty", None))

    def test_is_unique_nonexistent(self):
        facade = fixture_facade()
        with self.assertRaises(ResourceNonExistentException) as cm:
            facade.is_unique("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_is_unique_bad_type(self):
        resource_list = [
            ("primitive", ResourceType.PRIMITIVE),
            ("group", ResourceType.GROUP),
        ]
        for resource_id, resource_type in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceUnexpectedTypeException) as cm:
                    self.facade.is_unique(resource_id, None)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, None)
                self.assertEqual(cm.exception.resource_type, resource_type)
                self.assertEqual(
                    cm.exception.expected_types,
                    [ResourceType.CLONE, ResourceType.BUNDLE],
                )


class TestFacadeParentResources(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_get_parent_group_id(self):
        resource_list = [
            ("group-primitive_1", "group"),
            ("group-primitive_2", "group"),
            ("cloned_group-group-primitive_1", "cloned_group-group"),
        ]

        for resource_id, group_id in resource_list:
            with self.subTest(value=resource_id):
                result = self.facade.get_parent_group_id(resource_id, None)
                self.assertEqual(result, group_id)

    def test_get_parent_group_id_not_grouped(self):
        parent_id = self.facade.get_parent_group_id("primitive", None)
        self.assertEqual(parent_id, None)

    def test_get_parent_group_id_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.get_parent_group_id("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_get_parent_group_id_bad_type(self):
        resource_list = [
            ("group", ResourceType.GROUP),
            ("clone", ResourceType.CLONE),
            ("bundle", ResourceType.BUNDLE),
        ]

        for resource_id, resource_type in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceUnexpectedTypeException) as cm:
                    self.facade.get_parent_group_id(resource_id, None)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, None)
                self.assertEqual(cm.exception.resource_type, resource_type)
                self.assertEqual(
                    cm.exception.expected_types, [ResourceType.PRIMITIVE]
                )

    def test_get_parent_clone_id(self):
        resource_list = [
            ("clone-primitive", None, "clone"),
            ("clone_unique-primitive", None, "clone_unique"),
            ("clone_unique-primitive", "0", "clone_unique"),
            ("clone_unique-primitive", "1", "clone_unique"),
            ("cloned_group-group", None, "cloned_group"),
            ("cloned_group-group", "0", "cloned_group"),
            ("cloned_group-group-primitive_1", None, "cloned_group"),
        ]
        for resource_id, instance_id, expected_clone_id in resource_list:
            with self.subTest(value=resource_id):
                clone_id = self.facade.get_parent_clone_id(
                    resource_id, instance_id
                )
                self.assertEqual(
                    clone_id,
                    expected_clone_id,
                )

    def test_get_parent_clone_id_not_cloned(self):
        for resource_id in ["primitive", "group"]:
            with self.subTest(value=resource_id):
                result = self.facade.get_parent_clone_id(resource_id, None)
                self.assertEqual(result, None)

    def test_get_parent_clone_nonexistent(self):
        id_list = [
            ("nonexistent", None),
            ("primitive", "0"),
            ("clone-primitive", "42"),
            ("clone_unique-primitive", "42"),
        ]
        for resource_id, instance_id in id_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                with self.assertRaises(ResourceNonExistentException) as cm:
                    self.facade.get_parent_clone_id(resource_id, instance_id)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, instance_id)

    def test_get_parent_clone_id_bad_type(self):
        resource_list = [
            ("clone", ResourceType.CLONE),
            ("bundle", ResourceType.BUNDLE),
        ]

        for resource_id, resource_type in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceUnexpectedTypeException) as cm:
                    self.facade.get_parent_clone_id(resource_id, None)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, None)
                self.assertEqual(cm.exception.resource_type, resource_type)
                self.assertEqual(
                    cm.exception.expected_types,
                    [ResourceType.PRIMITIVE, ResourceType.GROUP],
                )

    def test_get_parent_bundle_id(self):
        parent_id = self.facade.get_parent_bundle_id("bundle-member", None)
        self.assertEqual(parent_id, "bundle")

    def test_get_parent_bundle_not_bundled(self):
        parent_id = self.facade.get_parent_bundle_id("primitive", None)
        self.assertEqual(parent_id, None)

    def test_get_parent_bundle_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.get_parent_bundle_id("nonexistent", None)

        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_get_parent_bundle_id_bad_type(self):
        resource_list = [
            ("clone", ResourceType.CLONE),
            ("bundle", ResourceType.BUNDLE),
        ]

        for resource_id, resource_type in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceUnexpectedTypeException) as cm:
                    self.facade.get_parent_bundle_id(resource_id, None)

                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, None)
                self.assertEqual(cm.exception.resource_type, resource_type)
                self.assertEqual(
                    cm.exception.expected_types, [ResourceType.PRIMITIVE]
                )


class TestGetIndexInGroup(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_ok(self):
        index_1 = self.facade.get_index_in_group("group-primitive_1", None)
        self.assertEqual(index_1, 0)

        index_2 = self.facade.get_index_in_group("group-primitive_2", None)
        self.assertEqual(index_2, 1)

    def test_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.get_index_in_group("nonexistent", None)
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_not_in_group(self):
        with self.assertRaises(ResourceNotInGroupException) as cm:
            self.facade.get_index_in_group("primitive", None)
        self.assertEqual(cm.exception.resource_id, "primitive")
        self.assertEqual(cm.exception.instance_id, None)


class TestFacadeGetMembers(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_all(self):
        resource_list = [
            ("group", None, ["group-primitive_1", "group-primitive_2"]),
            ("clone", None, ["clone-primitive"]),
            ("cloned_group", None, ["cloned_group-group"]),
            (
                "cloned_group-group",
                None,
                [
                    "cloned_group-group-primitive_1",
                    "cloned_group-group-primitive_2",
                ],
            ),
            (
                "cloned_group-group",
                "0",
                [
                    "cloned_group-group-primitive_1",
                    "cloned_group-group-primitive_2",
                ],
            ),
            ("bundle", None, ["bundle-member"]),
        ]

        for resource_id, instance_id, members in resource_list:
            with self.subTest(value=resource_id):
                result = self.facade.get_members(resource_id, instance_id)
                self.assertEqual(result, members)

    def test_nonexistent(self):
        resource_list = [("nonexistent", None), ("cloned_group-group", "42")]

        for resource_id, instance_id in resource_list:
            with self.subTest(value=resource_id):
                with self.assertRaises(ResourceNonExistentException) as cm:
                    self.facade.get_members(resource_id, instance_id)
                self.assertEqual(cm.exception.resource_id, resource_id)
                self.assertEqual(cm.exception.instance_id, instance_id)

    def test_bad_type(self):
        with self.assertRaises(ResourceUnexpectedTypeException) as cm:
            self.facade.get_members("primitive", None)

        self.assertEqual(cm.exception.resource_id, "primitive")
        self.assertEqual(cm.exception.instance_id, None)
        self.assertEqual(cm.exception.resource_type, ResourceType.PRIMITIVE)
        self.assertEqual(
            cm.exception.expected_types,
            [ResourceType.GROUP, ResourceType.CLONE, ResourceType.BUNDLE],
        )


def fixture_bundle_facade() -> ResourcesStatusFacade:
    return ResourcesStatusFacade(
        [
            BundleStatusDto(
                "bundle",
                "podman",
                "",
                unique=True,
                maintenance=False,
                description=None,
                managed=True,
                failed=False,
                replicas=[
                    BundleReplicaStatusDto(
                        "0",
                        member=fixture_primitive_dto(
                            "bundle-member",
                            None,
                            node_names=["bundle-remote-0"],
                        ),
                        remote=fixture_primitive_dto(
                            "bundle-remote-0", None, node_names=["node1"]
                        ),
                        container=fixture_primitive_dto(
                            "bundle-container-0", None, node_names=["node1"]
                        ),
                        ip_address=None,
                    ),
                    BundleReplicaStatusDto(
                        "1",
                        member=fixture_primitive_dto(
                            "bundle-member",
                            None,
                            target_role=PCMK_STATUS_ROLE_STOPPED,
                            managed=False,
                            node_names=["bundle-remote-1"],
                        ),
                        remote=fixture_primitive_dto(
                            "bundle-remote-1", None, node_names=["node2"]
                        ),
                        container=fixture_primitive_dto(
                            "bundle-container-1", None, node_names=["node2"]
                        ),
                        ip_address=None,
                    ),
                ],
            )
        ]
    )


class TestFacadeGetNodes(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_all(self):
        resource_list = [
            ("primitive", None, ["node1"]),
            ("primitive_stopped", None, []),
            ("group", None, ["node1"]),
            ("clone", None, ["node1", "node2"]),
            ("clone-primitive", None, ["node1", "node2"]),
            ("clone_unique", None, ["node0", "node0", "node1", "node1"]),
            (
                "clone_unique-primitive",
                None,
                ["node0", "node0", "node1", "node1"],
            ),
            ("clone_unique-primitive", "1", ["node1"]),
            ("clone_unique-primitive", "2", ["node0"]),
            ("cloned_group", None, ["node1"]),
            ("cloned_group-group", None, ["node1"]),
            ("cloned_group-group-primitive_1", None, ["node1"]),
            ("bundle", None, ["node0", "node1"]),
            ("bundle-member", None, ["node0", "node1"]),
        ]

        for resource_id, instance_id, node_list in resource_list:
            with self.subTest(value=f"{resource_id}:{instance_id}"):
                result = self.facade.get_nodes(resource_id, instance_id)
                self.assertEqual(result, node_list)

    def test_multiple_instances_on_same_node(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", str(i), node_names=[f"node{i % 2}"]
                        )
                        for i in range(4)
                    ],
                )
            ]
        )

        nodes = facade.get_nodes("clone", None)

        self.assertEqual(nodes, ["node0", "node0", "node1", "node1"])

    def test_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.get_nodes("nonexistent", None)
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)


class TestFacadeCanHaveMultipleMembers(TestCase):
    def test_nonexistent(self):
        facade = ResourcesStatusFacade([])

        with self.assertRaises(ResourceNonExistentException) as cm:
            facade.can_have_multiple_members("nonexistent", None)
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_all(self):
        facade = fixture_facade()
        resource_id_list = [
            ("stonith", False),
            ("primitive", False),
            ("group", True),
            ("group-primitive_1", False),
            ("clone", False),
            ("clone-primitive", False),
            ("cloned_group", True),
            ("cloned_group-group", True),
            ("cloned_group-group-primitive_1", False),
            ("bundle", False),
            ("bundle-member", False),
        ]
        for resource_id, result in resource_id_list:
            with self.subTest(value=f"{resource_id}"):
                self.assertEqual(
                    facade.can_have_multiple_members(resource_id, None), result
                )


class TestFacadeCanHaveMultipleInstances(TestCase):
    def setUp(self):
        self.facade = fixture_facade()

    def test_nonexistent(self):
        with self.assertRaises(ResourceNonExistentException) as cm:
            self.facade.can_have_multiple_instances("nonexistent", None)
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_all(self):
        resource_id_list = [
            ("stonith", False),
            ("primitive", False),
            ("group", False),
            ("group-primitive_1", False),
            ("clone", True),
            ("clone-primitive", True),
            ("cloned_group", True),
            ("cloned_group-group", True),
            ("cloned_group-group-primitive_1", True),
            ("bundle", True),
            ("bundle-member", True),
        ]
        for resource_id, result in resource_id_list:
            with self.subTest(value=f"{resource_id}"):
                self.assertEqual(
                    self.facade.can_have_multiple_instances(resource_id, None),
                    result,
                )

    def test_instance_id(self):
        self.assertTrue(
            self.facade.can_have_multiple_instances("clone_unique", None)
        )
        self.assertTrue(
            self.facade.can_have_multiple_instances(
                "clone_unique-primitive", None
            )
        )
        self.assertFalse(
            self.facade.can_have_multiple_instances(
                "clone_unique-primitive", "0"
            )
        )


class TestFacadeIsState(TestCase):
    # pylint: disable=too-many-public-methods
    def test_nonexistent(self):
        facade = ResourcesStatusFacade([])

        with self.assertRaises(ResourceNonExistentException) as cm:
            facade.is_state("nonexistent", None, ResourceState.STARTED)
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)

    def test_primitive_all_states(self):
        facade = ResourcesStatusFacade(
            [fixture_primitive_dto("primitive", None)]
        )

        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.STARTED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.STOPPED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.PROMOTED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.STARTING)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.STOPPING)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.DISABLED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.ENABLED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.MANAGED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.FAILED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.ACTIVE)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.ORPHANED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.BLOCKED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.FAILURE_IGNORED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.PENDING)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.LOCKED_TO)
        )

    def test_primitive_disabled(self):
        facade = ResourcesStatusFacade(
            [
                fixture_primitive_dto(
                    "primitive", None, target_role=PCMK_ROLE_STOPPED
                )
            ]
        )

        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.DISABLED)
        )
        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.ENABLED)
        )

    def test_primitive_pending_role_attribute(self):
        for role in PCMK_STATUS_ROLES_PENDING:
            with self.subTest(value=role):
                facade = ResourcesStatusFacade(
                    [fixture_primitive_dto("primitive", None, role=role)]
                )

                self.assertTrue(
                    facade.is_state("primitive", None, ResourceState.PENDING)
                )

    def test_primitive_pending_pending_attribute(self):
        facade = ResourcesStatusFacade(
            [fixture_primitive_dto("primitive", None, pending="Monitoring")]
        )

        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.PENDING)
        )

    def test_primitive_locked_to(self):
        facade = ResourcesStatusFacade(
            [fixture_primitive_dto("primitive", None, locked_to="node2")]
        )

        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.LOCKED_TO)
        )

    def test_primitive_on_node(self):
        facade = ResourcesStatusFacade(
            [fixture_primitive_dto("primitive", None)]
        )

        self.assertTrue(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                expected_node_name="node1",
            )
        )
        self.assertFalse(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                expected_node_name="node42",
            )
        )

    def test_primitive_on_node_stopped(self):
        facade = ResourcesStatusFacade(
            [
                fixture_primitive_dto(
                    "primitive",
                    None,
                    role=PCMK_STATUS_ROLE_STOPPED,
                    active=False,
                    node_names=[],
                )
            ]
        )

        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.STOPPED)
        )
        self.assertFalse(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STOPPED,
                expected_node_name="node1",
            )
        )

    def test_group_all(self):
        facade = ResourcesStatusFacade(
            [
                fixture_group_dto(
                    "group",
                    None,
                    members=[
                        fixture_primitive_dto("primitive_1", None),
                        fixture_primitive_dto("primitive_2", None),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.STARTED))
        self.assertFalse(facade.is_state("group", None, ResourceState.STOPPED))
        self.assertFalse(facade.is_state("group", None, ResourceState.PROMOTED))
        self.assertFalse(
            facade.is_state("group", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(facade.is_state("group", None, ResourceState.STARTING))
        self.assertFalse(facade.is_state("group", None, ResourceState.STOPPING))
        self.assertFalse(facade.is_state("group", None, ResourceState.DISABLED))
        self.assertTrue(facade.is_state("group", None, ResourceState.ENABLED))
        self.assertTrue(facade.is_state("group", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state("group", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("group", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(facade.is_state("group", None, ResourceState.FAILED))
        self.assertTrue(facade.is_state("group", None, ResourceState.ACTIVE))
        self.assertFalse(facade.is_state("group", None, ResourceState.ORPHANED))
        self.assertFalse(facade.is_state("group", None, ResourceState.BLOCKED))
        self.assertFalse(
            facade.is_state("group", None, ResourceState.FAILURE_IGNORED)
        )
        self.assertFalse(facade.is_state("group", None, ResourceState.PENDING))
        self.assertFalse(
            facade.is_state("group", None, ResourceState.LOCKED_TO)
        )

    def test_group_empty(self):
        facade = ResourcesStatusFacade([fixture_group_dto("group", None)])

        self.assertTrue(facade.is_state("group", None, ResourceState.MANAGED))
        self.assertFalse(facade.is_state("group", None, ResourceState.STARTED))

    def test_group_members_different_state(self):
        facade = ResourcesStatusFacade(
            [
                fixture_group_dto(
                    "group",
                    None,
                    members=[
                        fixture_primitive_dto("primitive-1", None),
                        fixture_primitive_dto(
                            "primitive-2",
                            None,
                            role=PCMK_STATUS_ROLE_STOPPED,
                            target_role=PCMK_ROLE_STOPPED,
                            failed=True,
                            active=False,
                            managed=False,
                            node_names=[],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.MANAGED))
        self.assertFalse(facade.is_state("group", None, ResourceState.DISABLED))

        self.assertFalse(facade.is_state("group", None, ResourceState.STARTED))
        self.assertFalse(facade.is_state("group", None, ResourceState.STOPPED))
        self.assertFalse(facade.is_state("group", None, ResourceState.FAILED))
        self.assertFalse(facade.is_state("group", None, ResourceState.ACTIVE))

    def test_group_members_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_group_dto(
                    "group",
                    None,
                    members=[
                        fixture_primitive_dto("primitive-1", None),
                        fixture_primitive_dto(
                            "primitive-2",
                            None,
                            role=PCMK_STATUS_ROLE_STOPPED,
                            active=False,
                            managed=False,
                            node_names=[],
                        ),
                    ],
                )
            ]
        )

        self.assertFalse(facade.is_state("group", None, ResourceState.STARTED))
        self.assertFalse(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "group",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "group",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "group",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_group_on_node(self):
        facade = ResourcesStatusFacade(
            [
                fixture_group_dto(
                    "group",
                    None,
                    members=[
                        fixture_primitive_dto("primitive-1", None),
                        fixture_primitive_dto(
                            "primitive-2",
                            None,
                            managed=False,
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.STARTED))
        self.assertTrue(
            facade.is_state(
                "group", None, ResourceState.STARTED, expected_node_name="node1"
            )
        )
        self.assertFalse(
            facade.is_state(
                "group", None, ResourceState.STARTED, expected_node_name="node2"
            )
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "group", None, ResourceState.MANAGED, expected_node_name="node1"
            )
        )

    def test_group_on_node_members_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_group_dto(
                    "group",
                    None,
                    members=[
                        fixture_primitive_dto("primitive-1", None),
                        fixture_primitive_dto(
                            "primitive-2",
                            None,
                            managed=False,
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("group", None, ResourceState.STARTED))
        self.assertTrue(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                expected_node_name="node1",
                members_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                expected_node_name="node1",
                members_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )

        self.assertTrue(
            facade.is_state(
                "group",
                None,
                ResourceState.STARTED,
                expected_node_name="node2",
                members_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_clone_bundle_in_status(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto("clone"),
                BundleStatusDto(
                    "bundle", "podman", "", True, False, None, True, False, []
                ),
            ]
        )
        with self.assertRaises(NotImplementedError):
            facade.is_state("clone", None, ResourceState.STARTED)

    def test_clone_all(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto(
                            "primitive", None, node_names=[node]
                        )
                        for node in ["node1", "node2"]
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.STARTED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.STOPPED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.PROMOTED))
        self.assertFalse(
            facade.is_state("clone", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(facade.is_state("clone", None, ResourceState.STARTING))
        self.assertFalse(facade.is_state("clone", None, ResourceState.STOPPING))
        self.assertFalse(facade.is_state("clone", None, ResourceState.DISABLED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.ENABLED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state("clone", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("clone", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(facade.is_state("clone", None, ResourceState.FAILED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.ACTIVE))
        self.assertFalse(facade.is_state("clone", None, ResourceState.ORPHANED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.BLOCKED))
        self.assertFalse(
            facade.is_state("clone", None, ResourceState.FAILURE_IGNORED)
        )
        self.assertFalse(facade.is_state("clone", None, ResourceState.PENDING))
        self.assertFalse(
            facade.is_state("clone", None, ResourceState.LOCKED_TO)
        )

    def test_empty_clone(self):
        facade = ResourcesStatusFacade([fixture_clone_dto("clone")])

        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.STARTED))

    def test_clone_orphans(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto("primitive", None),
                        fixture_primitive_dto(
                            "primitive",
                            None,
                            orphaned=True,
                            role=PCMK_STATUS_ROLE_STOPPED,
                            active=False,
                            node_names=[],
                        ),
                    ],
                ),
            ]
        )

        self.assertFalse(facade.is_state("clone", None, ResourceState.STOPPED))

    def test_clone_orphans_empty(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto("primitive", None, orphaned=True)
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.STARTED))

    def test_clone_members_different_state(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto(
                            "primitive", None, node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            None,
                            role=PCMK_STATUS_ROLE_STOPPED,
                            target_role=PCMK_ROLE_STOPPED,
                            failed=True,
                            active=False,
                            managed=False,
                            node_names=[],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.STARTED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.STOPPED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.DISABLED))
        self.assertFalse(facade.is_state("clone", None, ResourceState.FAILED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.ACTIVE))

    def test_clone_instances_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            role=PCMK_STATUS_ROLE_STOPPED,
                            managed=False,
                            node_names=[],
                            active=False,
                        ),
                    ],
                )
            ]
        )

        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_clone_on_node(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            managed=False,
                            node_names=["node2"],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                expected_node_name="node1",
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                expected_node_name="node2",
            )
        )

        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                expected_node_name="node1",
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                expected_node_name="node2",
            )
        )

    def test_clone_on_node_instances_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            managed=False,
                            node_names=["node2"],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.UNMANAGED,
                expected_node_name="node2",
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.UNMANAGED,
                expected_node_name="node2",
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )

    def test_clone_group_member(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_group_dto(
                            "group",
                            "0",
                            members=[
                                fixture_primitive_dto(
                                    "primitive", None, node_names=["node1"]
                                )
                            ],
                        ),
                        fixture_group_dto(
                            "group",
                            "1",
                            members=[
                                fixture_primitive_dto(
                                    "primitive",
                                    None,
                                    role=PCMK_STATUS_ROLE_STOPPED,
                                    active=False,
                                    managed=False,
                                    node_names=[],
                                )
                            ],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.STARTED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.STOPPED))
        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))

    def test_clone_group_member_instances_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_group_dto(
                            "group",
                            "0",
                            members=[
                                fixture_primitive_dto(
                                    "primitive", None, node_names=["node1"]
                                )
                            ],
                        ),
                        fixture_group_dto(
                            "group",
                            "1",
                            managed=False,
                            members=[
                                fixture_primitive_dto(
                                    "primitive",
                                    None,
                                    role=PCMK_STATUS_ROLE_STOPPED,
                                    active=False,
                                    managed=False,
                                    node_names=[],
                                )
                            ],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.STARTED))
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_clone_group_member_quantifiers(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_group_dto(
                            "group",
                            "0",
                            members=[
                                fixture_primitive_dto(
                                    "primitive", None, node_names=[node]
                                )
                                for node in ["node1", "node2"]
                            ],
                        ),
                        fixture_group_dto(
                            "group",
                            "1",
                            managed=False,
                            members=[
                                fixture_primitive_dto(
                                    "primitive",
                                    None,
                                    role=PCMK_STATUS_ROLE_STOPPED,
                                    active=False,
                                    managed=False,
                                    node_names=[],
                                ),
                                fixture_primitive_dto(
                                    "primitive2", None, node_names=["node2"]
                                ),
                            ],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.STARTED))
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(facade.is_state("clone", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.ANY,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertFalse(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertTrue(
            facade.is_state(
                "clone",
                None,
                ResourceState.MANAGED,
                members_quantifier=MoreChildrenQuantifierType.NONE,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_clone_instance(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            role=PCMK_STATUS_ROLE_STOPPED,
                            managed=False,
                            node_names=[],
                            active=False,
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(
            facade.is_state("primitive", "0", ResourceState.STARTED)
        )
        self.assertFalse(
            facade.is_state("primitive", "1", ResourceState.STARTED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.STARTED)
        )

        self.assertFalse(
            facade.is_state("primitive", "0", ResourceState.STOPPED)
        )
        self.assertTrue(
            facade.is_state("primitive", "1", ResourceState.STOPPED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.STOPPED)
        )

        self.assertTrue(
            facade.is_state("primitive", "0", ResourceState.MANAGED)
        )
        self.assertFalse(
            facade.is_state("primitive", "1", ResourceState.MANAGED)
        )
        self.assertTrue(
            facade.is_state("primitive", None, ResourceState.MANAGED)
        )

    def test_clone_one_instance_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[fixture_primitive_dto("primitive", None)],
                )
            ]
        )
        self.assertTrue(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )

    def test_clone_instance_orphans(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto("primitive", None),
                        fixture_primitive_dto(
                            "primitive",
                            None,
                            orphaned=True,
                            role=PCMK_STATUS_ROLE_STOPPED,
                            active=False,
                            node_names=[],
                        ),
                    ],
                ),
            ]
        )

        self.assertFalse(
            facade.is_state("primitive", None, ResourceState.STOPPED)
        )

    def test_clone_instance_orphans_empty(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    instances=[
                        fixture_primitive_dto("primitive", None, orphaned=True)
                    ],
                ),
            ]
        )

        with self.assertRaises(ResourceNonExistentException) as cm:
            facade.is_state("primitive", None, ResourceState.STARTED)
        self.assertEqual(cm.exception.resource_id, "primitive")
        self.assertEqual(cm.exception.instance_id, None)

    def test_clone_instance_instances_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            role=PCMK_STATUS_ROLE_STOPPED,
                            node_names=[],
                            active=False,
                        ),
                    ],
                )
            ]
        )

        self.assertFalse(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "primitive",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(
            facade.is_state(
                "primitive",
                None,
                ResourceState.PROMOTED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_clone_instance_on_node(self):
        facade = ResourcesStatusFacade(
            [
                fixture_clone_dto(
                    "clone",
                    unique=True,
                    instances=[
                        fixture_primitive_dto(
                            "primitive", "0", node_names=["node1"]
                        ),
                        fixture_primitive_dto(
                            "primitive",
                            "1",
                            managed=False,
                            node_names=["node2"],
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(
            facade.is_state(
                "primitive",
                None,
                ResourceState.MANAGED,
                expected_node_name="node1",
            )
        )
        self.assertFalse(
            facade.is_state(
                "primitive",
                None,
                ResourceState.MANAGED,
                expected_node_name="node2",
            )
        )

    def test_bundle_all(self):
        facade = fixture_bundle_facade()

        self.assertTrue(facade.is_state("bundle", None, ResourceState.STARTED))
        self.assertFalse(facade.is_state("bundle", None, ResourceState.STOPPED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.PROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.STARTING)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.STOPPING)
        )
        self.assertTrue(facade.is_state("bundle", None, ResourceState.DISABLED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.ENABLED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.FAILED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.ACTIVE))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.ORPHANED)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.BLOCKED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.FAILURE_IGNORED)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.PENDING))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.LOCKED_TO)
        )

    def test_bundle_no_member_all(self):
        facade = ResourcesStatusFacade(
            [
                BundleStatusDto(
                    "bundle",
                    "podman",
                    "",
                    True,
                    False,
                    None,
                    True,
                    False,
                    [
                        BundleReplicaStatusDto(
                            "0",
                            member=None,
                            remote=None,
                            container=fixture_primitive_dto(
                                "bundle-container-0", None, node_names=["node1"]
                            ),
                            ip_address=None,
                        ),
                        BundleReplicaStatusDto(
                            "1",
                            member=None,
                            remote=None,
                            container=fixture_primitive_dto(
                                "bundle-container-0",
                                None,
                                target_role=PCMK_STATUS_ROLE_STOPPED,
                                node_names=["node2"],
                            ),
                            ip_address=None,
                        ),
                    ],
                )
            ]
        )

        self.assertTrue(facade.is_state("bundle", None, ResourceState.STARTED))
        self.assertFalse(facade.is_state("bundle", None, ResourceState.STOPPED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.PROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.STARTING)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.STOPPING)
        )
        self.assertTrue(facade.is_state("bundle", None, ResourceState.DISABLED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.ENABLED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.FAILED))
        self.assertTrue(facade.is_state("bundle", None, ResourceState.ACTIVE))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.ORPHANED)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.BLOCKED))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.FAILURE_IGNORED)
        )
        self.assertFalse(facade.is_state("bundle", None, ResourceState.PENDING))
        self.assertFalse(
            facade.is_state("bundle", None, ResourceState.LOCKED_TO)
        )

    def test_bundle_instances_quantifier(self):
        facade = fixture_bundle_facade()

        self.assertTrue(facade.is_state("bundle", None, ResourceState.STARTED))
        self.assertTrue(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

        self.assertTrue(facade.is_state("bundle", None, ResourceState.MANAGED))
        self.assertFalse(
            facade.is_state(
                "bundle",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "bundle",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "bundle",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_bundle_on_node(self):
        facade = fixture_bundle_facade()

        self.assertTrue(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                expected_node_name="node1",
            )
        )
        self.assertTrue(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                expected_node_name="node2",
            )
        )
        self.assertFalse(
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                expected_node_name="bundle-remote-0",
            )
        )

    def test_bundle_member_all(self):
        facade = fixture_bundle_facade()

        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.STARTED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.STOPPED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.PROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.UNPROMOTED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.STARTING)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.STOPPING)
        )
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.DISABLED)
        )
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.ENABLED)
        )
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.MANAGED)
        )
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.UNMANAGED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.MAINTENANCE)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.FAILED)
        )
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.ACTIVE)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.ORPHANED)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.BLOCKED)
        )
        self.assertFalse(
            facade.is_state(
                "bundle-member", None, ResourceState.FAILURE_IGNORED
            )
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.PENDING)
        )
        self.assertFalse(
            facade.is_state("bundle-member", None, ResourceState.LOCKED_TO)
        )

    def test_bundle_member_quantifier(self):
        facade = fixture_bundle_facade()

        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.MANAGED)
        )
        self.assertFalse(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ALL,
            )
        )
        self.assertTrue(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.ANY,
            )
        )
        self.assertFalse(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                instances_quantifier=MoreChildrenQuantifierType.NONE,
            )
        )

    def test_bundle_member_on_node(self):
        facade = fixture_bundle_facade()
        self.assertTrue(
            facade.is_state("bundle-member", None, ResourceState.MANAGED)
        )
        self.assertTrue(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                expected_node_name="node1",
            )
        )
        self.assertFalse(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                expected_node_name="node2",
            )
        )

        self.assertFalse(
            facade.is_state(
                "bundle-member",
                None,
                ResourceState.MANAGED,
                expected_node_name="bundle-remote-0",
            )
        )

    def test_bad_members_quantifier(self):
        facade = ResourcesStatusFacade(
            [
                fixture_primitive_dto("primitive", None),
                fixture_clone_dto(
                    "clone",
                    instances=[fixture_primitive_dto("cloned_primitive", None)],
                ),
            ]
        )

        resource_list = ["primitive", "clone", "cloned_primitive"]
        for resource_id in resource_list:
            with (
                self.subTest(value=resource_id),
                self.assertRaises(MembersQuantifierUnsupportedException),
            ):
                facade.is_state(
                    resource_id,
                    None,
                    ResourceState.STARTED,
                    members_quantifier=MoreChildrenQuantifierType.ALL,
                )

    def test_bad_members_quantifier_bundle(self):
        facade = fixture_bundle_facade()
        with self.assertRaises(MembersQuantifierUnsupportedException):
            facade.is_state(
                "bundle",
                None,
                ResourceState.STARTED,
                members_quantifier=MoreChildrenQuantifierType.ALL,
            )

    def test_bad_instances_quantifier(self):
        facade = fixture_facade()

        resource_list = ["primitive", "group"]
        for resource_id in resource_list:
            with (
                self.subTest(value=resource_id),
                self.assertRaises(InstancesQuantifierUnsupportedException),
            ):
                facade.is_state(
                    resource_id,
                    None,
                    ResourceState.STARTED,
                    instances_quantifier=MoreChildrenQuantifierType.ALL,
                )


class TestStateExactValue(TestCase):
    def test_locked_to(self):
        facade = ResourcesStatusFacade(
            [fixture_primitive_dto("primitive", None, locked_to="node1")]
        )

        result = facade.is_state_exact_value(
            "primitive", None, ResourceState.LOCKED_TO, "node1"
        )
        self.assertTrue(result)

        bad_result = facade.is_state_exact_value(
            "primitive", None, ResourceState.LOCKED_TO, "node42"
        )
        self.assertFalse(bad_result)

    def test_nonexistent(self):
        facade = ResourcesStatusFacade([])

        with self.assertRaises(ResourceNonExistentException) as cm:
            facade.is_state_exact_value(
                "nonexistent", None, ResourceState.PENDING, "42"
            )
        self.assertEqual(cm.exception.resource_id, "nonexistent")
        self.assertEqual(cm.exception.instance_id, None)
